/*
 * IRC - Internet Relay Chat, ircd/m_svsnick.c
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: m_svsnick.c, Valterius Exp $
 */

/*
 * m_functions execute protocol messages on this server:
 *
 *    cptr    is always NON-NULL, pointing to a *LOCAL* client
 *            structure (with an open socket connected!). This
 *            identifies the physical socket where the message
 *            originated (or which caused the m_function to be
 *            executed--some m_functions may call others...).
 *
 *    sptr    is the source of the message, defined by the
 *            prefix part of the message if present. If not
 *            or prefix not found, then sptr==cptr.
 *
 *            (!IsServer(cptr)) => (cptr == sptr), because
 *            prefixes are taken *only* from servers...
 *
 *            (IsServer(cptr))
 *                    (sptr == cptr) => the message didn't
 *                    have the prefix.
 *
 *                    (sptr != cptr && IsServer(sptr) means
 *                    the prefix specified servername. (?)
 *
 *                    (sptr != cptr && !IsServer(sptr) means
 *                    that message originated from a remote
 *                    user (not local).
 *
 *            combining
 *
 *            (!IsServer(sptr)) means that, sptr can safely
 *            taken as defining the target structure of the
 *            message in this server.
 *
 *    *Always* true (if 'parse' and others are working correct):
 *
 *    1)      sptr->from == cptr  (note: cptr->from == cptr)
 *
 *    2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr
 *            *cannot* be a local connection, unless it's
 *            actually cptr!). [MyConnect(x) should probably
 *            be defined as (x == x->from) --msa ]
 *
 *    parc    number of variable parameter strings (if zero,
 *            parv is allowed to be NULL)
 *
 *    parv    a NULL terminated list of parameter pointers,
 *
 *                    parv[0], sender (prefix string), if not present
 *                            this points to an empty string.
 *                    parv[1]...parv[parc-1]
 *                            pointers to additional parameters
 *                    parv[parc] == NULL, *always*
 *
 *            note:   it is guaranteed that parv[0]..parv[parc-1] are all
 *                    non-NULL pointers.
 */
#include "config.h"
#include "channel.h"
#include "client.h"
#include "handlers.h"
#include "hash.h"
#include "ircd.h"
#include "ircd_reply.h"
#include "ircd_string.h"
#include "msg.h"
#include "numeric.h"
#include "numnicks.h"
#include "s_conf.h"
#include "s_debug.h"
#include "s_user.h"
#include "send.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

static int check_nickname(char* nick)
{
  char* ch  = nick;
  char* end = ch + NICKLEN;
  assert(0 != ch);

 /* can't insert an `|' caracter in first time without bug --Valterius */
  if (*ch == '-' || IsDigit(*ch) || *ch == '|')        /* first character in [0..9-] */
    return 0;

  for ( ; (ch < end) && *ch; ++ch)
    if (!IsNickChar(*ch))
      break;

  *ch = '\0';

  return (ch - nick);
}

static int do_svsnick(struct Client* sptr, struct Client* acptr, char* newnick)
{
	if (MyUser(acptr)) { 
		struct Membership *member;
		char nick[NICKLEN + 2];
		char* arg;
		char* s;
		arg = newnick;
		if (strlen(arg) > NICKLEN)
		arg[NICKLEN] = '\0';
		if ((s = strchr(arg, '~')))
		*s = '\0';
		strcpy(nick, arg);
		for (member = (cli_user(acptr))->channel; member; member = member->next_channel)
			ClearBanValid(member);
		sendcmdto_common_channels_butone(acptr, CMD_NICK, NULL, ":%s", nick);
		add_history(acptr, 1);
		sendcmdto_serv_butone(acptr, CMD_NICK, acptr, "%s %Tu", nick, cli_lastnick(acptr));
		if ((cli_name(acptr))[0])
			hRemClient(acptr);
		strcpy(cli_name(acptr), nick);
		hAddClient(acptr);
	} 
	else {
		sendcmdto_one(sptr, MSG_SVSNICK, TOK_SVSNICK, acptr, "%C :%s", acptr, newnick);
	}
	return 0;
}

/*
 *
 * mo_svsnick - opers
 * parv[0] = exp
 * parv[1] = nick
 * parv[2] = new nick
 *
 * --Valterius
 *
 */
int mo_svsnick(struct Client* cptr, struct Client* sptr, int parc, char* parv[])
{
	struct User*     user = cli_user(sptr);
	struct Client* acptr = NULL;
	struct Client* acptr2 = NULL;

	if (parc < 3)
		return(need_more_params(sptr, "SVSNICK"));

	if (!check_nickname(parv[2])) {
		send_reply(sptr, ERR_ERRONEUSNICKNAME, parv[2]);
		return 0;
	}
	if(!(acptr = FindClient(parv[1])))
		return send_reply(sptr, ERR_NOSUCHNICK, parv[1]);

	if (IsChannelService(acptr)) /* Cannot svsnick services --Valterius */
		return send_reply(sptr, ERR_NOPRIVILEGES);
    
	if (cli_name(acptr) == parv[2])
		return 0;
	if ((acptr2 = FindClient(parv[2]))) {
		send_reply(sptr, ERR_NICKNAMEINUSE, cli_name(acptr2));
		return 0;
	}
	/* global server notice --Valterius */
	sendto_allops(&me, SNO_HACK4, "HACK: %s SVSNICK %s %s", parv[0], parv[1], parv[2]);
	return do_svsnick(sptr, acptr, parv[2]);
}

/*
 *
 * ms_svsnick - servers
 * parv[0] = exp
 * parv[1] = num
 * parv[2] = new nick
 *
 * --Valterius
 *
 */
int ms_svsnick(struct Client* cptr, struct Client* sptr, int parc, char* parv[])
{
	struct Client* acptr;
	struct Client* acptr2;
	char* param[3];
	char* tmp;

	if (parc < 3) {
		protocol_violation(sptr,"Too few arguments for SVSNICK");
		return need_more_params(sptr, "SVSNICK");
	}
	if ((acptr = findNUser(parv[1])) && !IsUser(acptr)) {
		protocol_violation(sptr,"Invalid numeric for SVSNICK");
		return 0;
	}
	if (!check_nickname(parv[2])) {
		protocol_violation(sptr,"Invalid nickname for SVSNICK");
		return 0;
	}

	if((acptr2 = FindClient(parv[2]))) {
//		++ServerStats->is_kill;
		SetFlag(cptr, FLAG_KILLED);
		exit_client(cptr, acptr2, &me, "Killed (SVSNICK Collision)");
	}

	if (cli_name(acptr) == parv[2])
		return 0;

	return do_svsnick(sptr, acptr, parv[2]);
}
